/******************************************************************

        MUSCLE SmartCard Development ( http://www.linuxnet.com )
            Title  : csp-cryptoflex.c
            Package: RSA Smartcard PAM module
            Author : David Corcoran
            Date   : 4/12/00
            Purpose: This provides some services for the Cryptoflex Card.

********************************************************************/
#include <includes.h>
#include <ssh.h>
#include <gmp.h>
#include <xmalloc.h>
#include <rsa.h>
#include <stdio.h>
#include <winscard.h>

UCHAR ACCESS_AUT0_KEY[8] = {0x4D, 0x49, 0x43, 0x52, 0x4F, 0x53, 0x46, 0x54};
DWORD ACCESS_AUT0_SIZE = 8;
UCHAR ACCESS_CHV0_VAL[8] = {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30};
DWORD ACCESS_CHV0_SIZE = 8;

LONG SCSSelectFile( SCARDHANDLE hLCard, UCHAR ucF1, UCHAR ucF2 ) {

  LONG rv;
  UCHAR ucCommand[50];
  UCHAR ucRecv[50];
  DWORD dwLength;
  SCARD_IO_REQUEST sRecvPci;

  dwLength = 2;
  sRecvPci.dwProtocol = SCARD_PROTOCOL_T0;

  ucCommand[0] = 0xC0; ucCommand[1] = 0xA4; ucCommand[2] = 0x00;
  ucCommand[3] = 0x00; ucCommand[4] = 0x02; ucCommand[5] = ucF1;
  ucCommand[6] = ucF2;

  rv = SCardTransmit( hLCard, SCARD_PCI_T0, ucCommand, 7, &sRecvPci, 
		      ucRecv, &dwLength );
  
  if ( dwLength == 2 && ucRecv[0] == 0x61 && rv == SCARD_S_SUCCESS ) {
    return rv;
  } else {
    printf("Invalid FILE\n");
    return SCARD_E_INVALID_VALUE;
  }
}

LONG SCSVerifyKey ( SCARDHANDLE hLCard, DWORD dwKSize, PUCHAR pucKey ) {

  LONG rv;
  UCHAR ucCommand[50];
  UCHAR ucRecv[50];
  DWORD dwLength;
  SCARD_IO_REQUEST sRecvPci;

  sRecvPci.dwProtocol = SCARD_PROTOCOL_T0;
  dwLength            = 2;

  ucCommand[0] = 0xF0; ucCommand[1] = 0x2A; ucCommand[2] = 0x00;
  ucCommand[3] = 0x01; ucCommand[4] = dwKSize;

  memcpy( &ucCommand[5], pucKey, dwKSize );

  rv = SCardTransmit( hLCard, SCARD_PCI_T0, ucCommand, dwKSize + 5, &sRecvPci, 
		      ucRecv, &dwLength );

  if ( dwLength == 2 && ucRecv[0] == 0x90 && rv == SCARD_S_SUCCESS ) {
    return rv;
  } else {
    printf("Invalid KEY AUT\n");
    return SCARD_E_INVALID_VALUE;
  }
}

LONG SCSVerifyChv ( SCARDHANDLE hLCard, DWORD dwCSize, PUCHAR pucChv ) {

  LONG rv;
  UCHAR ucCommand[50];
  UCHAR ucRecv[50];
  DWORD dwLength;
  SCARD_IO_REQUEST sRecvPci;

  sRecvPci.dwProtocol = SCARD_PROTOCOL_T0;
  dwLength            = 2;

  ucCommand[0] = 0xC0; ucCommand[1] = 0x20; ucCommand[2] = 0x00;
  ucCommand[3] = 0x01; ucCommand[4] = dwCSize;

  memcpy( &ucCommand[5], pucChv, dwCSize );

  rv = SCardTransmit( hLCard, SCARD_PCI_T0, ucCommand, dwCSize + 5, &sRecvPci, 
		      ucRecv, &dwLength );
  
  if ( dwLength == 2 && ucRecv[0] == 0x90 && rv == SCARD_S_SUCCESS ) {
    return rv;
  } else {
    printf("Invalid CHV AUT\n");
    return SCARD_E_INVALID_VALUE;
  }
}

LONG SCSRSASignData( SCARDHANDLE hLCard, PUCHAR pucIBuffer, 
		     PUCHAR pucOBuffer, LPDWORD dwSize ) {
  
  LONG rv;
  UCHAR ucCommand[200];
  UCHAR ucReverse[200];
  UCHAR ucRecv[200];
  DWORD dwLength;
  SCARD_IO_REQUEST sRecvPci;
  int i, j;

  rv = SCSSelectFile(hLCard, 0x3F, 0x11 );

  ucCommand[0] = 0xC0;
  ucCommand[1] = 0x88;
  ucCommand[2] = 0x00;
  ucCommand[3] = 0x00;
  ucCommand[4] = 0x80;

  j=0;

  for (i=0x80 - 1; i >= 0; i--) {
    ucReverse[j++] = pucIBuffer[i];
  }
  

  memcpy( &ucCommand[5], ucReverse, *dwSize ); 
  sRecvPci.dwProtocol = SCARD_PROTOCOL_T0;
  rv = SCardTransmit( hLCard, SCARD_PCI_T0, ucCommand, *dwSize+5, 
		      &sRecvPci, pucOBuffer, dwSize );

  if ( pucOBuffer[0] != 0x61 || pucOBuffer[1] != 0x80 || 
       rv != SCARD_S_SUCCESS )
    {
      return SCARD_E_INVALID_VALUE;
    }

  
  ucCommand[0] = 0xC0;
  ucCommand[1] = 0xC0;
  ucCommand[2] = 0x00;
  ucCommand[3] = 0x00;
  ucCommand[4] = 0x80;
  
  *dwSize = 0x80;
  
  sRecvPci.dwProtocol = SCARD_PROTOCOL_T0;
  rv = SCardTransmit( hLCard, SCARD_PCI_T0, ucCommand, 5, &sRecvPci,
		      ucReverse, dwSize );

  j=0;

  for (i=0x80 - 1; i >= 0; i--) {
    pucOBuffer[j++] = ucReverse[i];
  } 

  return rv;
}

#define MAX_KEY_FILE_SIZE 1024
#define PRV_KEY_SIZE 64*6
#define PRV_KEY_FID0 0x00
#define PRV_KEY_FID1 0x12
#define NUM_RSA_KEY_ELEMENTS 5
#define RSA_BIT_LEN 1024
#define key_number 0x10
#define key_type 0xc8 /* key type 0xc8 (1024 bit RSA private) */
#define KEY_FILE_HEADER_SIZE 8
#define MAX_APDU_SIZE 0x10

LONG SCSLoadRsaKPriv ( SCARDHANDLE hCard, PUCHAR GCRTKeyVal, 
		      DWORD GCRTKeySize ) {
  UCHAR ucCommand[1000];
  UCHAR ucRecv[500];
  UCHAR ucTemp[128];
  DWORD dwLength;
  SCARD_IO_REQUEST sRecvPci;

  int r1, r2, rv, argc = 0, i, j; 
  int tmp, offset=0, size, crtspot;
  unsigned char *token; 
  int key_num[2];
  unsigned char data[MAX_KEY_FILE_SIZE];
  
  /* select 00.12 */
  rv = SCSSelectFile(hCard, 0x3F, 0x11 );
  rv = SCSSelectFile(hCard, 0x00, 0x12 );
  if ( rv != SCARD_S_SUCCESS ) {
    printf("Creating Private Key File 0x0012\n\n");
  }

  /* Start the graph */
  printf("Writing RSA Private Key: 0x0012\n");
  printf("[0 -------- 50 --------- 100]\n");
  
  fputs("[*", stdout);
  fflush(stdout);
  
  /* select 3f.00 */
  
  rv = SCSSelectFile(hCard, 0x3f, 0);
  if (rv != SCARD_S_SUCCESS) return rv;
  
  fputs("*", stdout);
  fflush(stdout);
  
  rv = SCSVerifyKey ( hCard, ACCESS_AUT0_SIZE, ACCESS_AUT0_KEY );
  if (rv != SCARD_S_SUCCESS) return rv;
  
  rv = SCSVerifyChv ( hCard, ACCESS_CHV0_SIZE, ACCESS_CHV0_VAL );
  
  fputs("*", stdout);
  fflush(stdout);
  
  if (rv != SCARD_S_SUCCESS) return rv;
  
  /* select 00.12 */
  rv = SCSSelectFile(hCard, 0x3f, 0x11 );
  rv = SCSSelectFile(hCard, 0x00, 0x12 );
  fputs("*", stdout);
  fflush(stdout);
  
  if (rv != SCARD_S_SUCCESS ) {
    
    /* create private key file */
    data[0] = (PRV_KEY_SIZE + 16) / 256; /* size, upper byte */
    data[1] = (PRV_KEY_SIZE + 16) % 256; /* size, lower byte */
    data[2] = PRV_KEY_FID0;   /* FID, upper */
    data[3] = PRV_KEY_FID1;   /* FID, lower */
    data[4] = 0x03;           /* file type = 2 (binary file) */
    data[5] = 0x01;           /* status = 1 */
    data[6] = data[7] = 0x00; /* record related */
    data[8] = 0xff;           /* ACL can do everything with AUT0 */
    for (i = 9; i < 16; i++ ) {
      data[i] = 0x00;         /* ACL : cannot do anything without AUT0 */
    }
    
    ucCommand[0] = 0xC0; ucCommand[1] = 0xe0; ucCommand[2] = 0x00;
    ucCommand[3] = 0x00; ucCommand[4] = 0x10;
    dwLength = 2;
    
    memcpy( &ucCommand[5], data, 0x10 );
    
    rv = SCardTransmit( hCard, SCARD_PCI_T0, ucCommand, 0x10 + 5, &sRecvPci, 
			ucRecv, &dwLength );
    
    r1 = ucRecv[0]; r2 = ucRecv[1];
    
    if (r1 != 0x90 && r1 != 0x61) {
      /* error */
      printf("Error: Creating File %02x.%02x: ",
	     PRV_KEY_FID0, PRV_KEY_FID1);
      // print_r1r2(r1, r2);
      return -1; 
    }
  }
  
  /* burn the key */
  data[0] = 0x01;       /* key size, I guess */
  data[1] = 0x43;       /* key size, I guess */
  data[2] = 0x01;       /* key number */
  
  memcpy( &data[3], GCRTKeyVal, 64*5 ); 
  
  /* now send this to the card */
  /* select private key file */
  rv = SCSSelectFile(hCard, 0x00, 0x12);
  fputs("*", stdout);
  fflush(stdout);
  
  if (rv != SCARD_S_SUCCESS) return rv;
  
  /* update binary */
  size = (64*5)+3;
  
  for (i = 0; i < size; i += MAX_APDU_SIZE) {
    int send_size;
    
    /* compute the size to be sent */
    if (size - i > MAX_APDU_SIZE) send_size = MAX_APDU_SIZE;
    else send_size = size - i;
    
    ucCommand[0] = 0xC0;  ucCommand[1] = 0xd6; ucCommand[2] = i/256;
    ucCommand[3] = i%256; ucCommand[4] = send_size;
    dwLength = 2;
    
    memcpy( &ucCommand[5], data + i, send_size );
    
    SCardTransmit( hCard, SCARD_PCI_T0, ucCommand, send_size+5, &sRecvPci, 
		   ucRecv, &dwLength );
    fputs("*", stdout);
    fflush(stdout);
    
    r1 = ucRecv[0]; r2 = ucRecv[1];
    
    if (r1 != 0x90 && r1 != 0x61 && dwLength != 2) {
      /* error */
      printf("Error: Updating Binary %02x.%02x: ", 0x00, 0x12);
      return -1; 
    } 
  }
  
  printf("] Done. \n\n");
  return 0; 
}  

void FormatMPInt (PUCHAR pucBuffer, MP_INT * mpint)  {
  int i0, j0, k0;
  
  k0 = 0;
  
  for (i0=abs(mpint->_mp_size)-1; i0>=0; i0--) {
    for (j0=3; j0>=0; j0--) {
      pucBuffer[k0++] = (mpint->_mp_d[i0]>>(j0*8)) & 0xff;	
    }
  }
}

void ComputeAndFormatDmodX (PUCHAR pucBuffer, MP_INT * dx, MP_INT * x, 
			    MP_INT * d ) {
  mpz_init(dx);
  mpz_sub_ui(dx, x, 1);
  mpz_mod(dx, d, dx);
  FormatMPInt ( pucBuffer, dx );
  mpz_clear(dx);
  return;
}

void FormatPublicKey (RSAPublicKey * key, PUCHAR pucBuffer) {
  MP_INT dp, dq;

  FormatMPInt (&pucBuffer[0], &key->e);
  FormatMPInt (&pucBuffer[4], &key->n);
}

void FormatPrivateKey (RSAPrivateKey * key, PUCHAR pucBuffer)
{
  MP_INT dp, dq;
  UCHAR pucKey[1000];
  int i, p;

  p=0;

  FormatMPInt (&pucBuffer[64*0], &key->q);
  for (i=(64*1)-1; i >=64*0; i--) pucKey[p++] = pucBuffer[i];

  FormatMPInt (&pucBuffer[64*1], &key->p);
  for (i=(64*2)-1; i >=64*1; i--) pucKey[p++] = pucBuffer[i];

  FormatMPInt (&pucBuffer[64*2], &key->u);
  for (i=(64*3)-1; i >=64*2; i--) pucKey[p++] = pucBuffer[i];

  /* Compute dq = d mod q-1. */
  ComputeAndFormatDmodX (&pucBuffer[64*3], &dq, &key->q, &key->d);
  for (i=(64*4)-1; i >=64*3; i--) pucKey[p++] = pucBuffer[i];

  /* Compute dp = d mod p-1. */
  ComputeAndFormatDmodX (&pucBuffer[64*4], &dp, &key->p, &key->d);
  for (i=(64*5)-1; i >=64*4; i--) pucKey[p++] = pucBuffer[i];

  FormatMPInt (&pucBuffer[65*5], &key->n);
  for (i=(64*6)-1; i >=64*5; i--) pucKey[p++] = pucBuffer[i];

  memcpy( pucBuffer, pucKey, 64*6 );

}


LONG SCSFormatRsaKPriv ( RSAPrivateKey *Kpriv, PUCHAR GCRTKeyVal ) {
  FormatPrivateKey( Kpriv, GCRTKeyVal );
}
